%%%--------------------------------------------------------------------- 
%%% Copyright (C) 2008  Erik G. Burrows
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%
%%%--------------------------------------------------------------------- 
%%% Description module ais_pgsql_log
%%%--------------------------------------------------------------------- 
%%% Log parsed AIS messages to a PostgreSQL database.
%%%
%%%--------------------------------------------------------------------- 
%%% Exports
%%%--------------------------------------------------------------------- 
%%% @spec start_link(Name, [Host, Username, Password, Database]) -> {ok, PID}
%%%
%%%     gen_server wrapper for init() function. Instantiates the
%%%     module with the specified name, and passes in startup parameters
%%%
%%%     Name = Name of this instance of this module
%%%     Host = PostgreSQL hostname
%%%     Username = PostgreSQL account name
%%%     Password = Password to PostgreSQL database
%%%     Database = Database name
%%%
%%% @spec input(Name, Message) -> {ok}
%%%     Log a parsed AIS message to the database
%%%     Name = Name of module instance
%%%     Message = Parsed message to log
%%%
%%%--------------------------------------------------------------------- 
%%% Requirements
%%%--------------------------------------------------------------------- 
%%%
%%% epsql PostgreSQL driver must be in pathz. See:
%%%     http://www.erlangatwork.com/2009/01/erlang-and-postgresql.html
%%%
%%%--------------------------------------------------------------------- 
%%% Revision History
%%%--------------------------------------------------------------------- 
%%% Revision 0.0.1 Date 2009/02/20 Author Erik Burrows erik@erikburrows.com
%%%     Initial revision, basic implementation of modular GNUradio-ish
%%%     structure.
%%%
-module(ais_pgsql_log).


%% gen_server is used to stop/start module
-behavior(gen_server).


%% This module sends but does not receive UDP packets, so it's a sink
-behavior(gen_sink).


-include_lib("common_records.hrl").
-include_lib("ais_patterns.hrl").


%% Public exported functions.
%%   input/3: Send a message to the remote host
%%   init, stop, handle..: gen_server required public functions
%%
-export([
    start_link/2,
    input/5,

    init/1,
    stop/1,
    handle_call/3,
    handle_cast/2,
    handle_info/2,
    terminate/2,
    code_change/3
    ]).


%% Module input/output signatures, as required by manager program
%%
-inputs([{"NMEA", nmea_type}]).
-outputs([]).


%% Data Type: state
%% Where:
%%   name = Name of this instance of this module (ex. Chartplotter)
%%   host = Remote host to connect to
%%   port = TCP port on remote host
%%   socket = Opened client socket
%%
-record(state, {name, conn}).


%% @spec input(Name, InputName, Message) -> {ok}
%%
%% Send a message to the module, to be sent to the remote host.
%%
%% Name = Name of this instance of this module
%% InputName = Name of the input to send to
%%
input(Name, InputName, OutputModule, OutputName, Message) -> 
    gen_server:call({global, Name}, {input, InputName, OutputModule, OutputName, Message}).


%% @spec start_link(Name, [Host, Username, Password, Database]) -> {ok, PID}
%%
%% gen_server wrapper for init() function. Instantiates the
%% module with the specified name, and passes in startup parameters
%%
%% Host = PostgreSQL host
%% Username = User name
%% Password = Password
%% Database = Database name
%%
start_link(Name, [Host, Username, Password, Database]) ->
    gen_server:start_link({global, Name}, ?MODULE, [Name | [[Host, Username, Password, Database]]], []).


%% @spec init([Name, [Host, Username, Password, Database]]) -> {ok, State}
%%
%% Startup function, called only by start_link RPC wrapper.
%%
init([Name, [Host, Username, Password, Database]]) ->
    io:format("~p starting\n", [?MODULE]),

    {ok, C} = pgsql:connect(Host, Username, Password, [{database, Database}]),

    {ok, #state{name=Name, conn=C}}.


%% @spec handle_call({name, Arg, ...}) -> {reply, ok, Arg, ...}
%%
%% Accept message to input named NMEA
%%
handle_call({input, "NMEA", _OutputModule, _OutputName, Message}, _From, State) ->
    #state{conn=C} = State,

    ok = log_message(C, Message),

    {reply, ok, State}.

% @spec handle_info(Info, State) -> {noreply, State}
%
handle_info(_Info, State) ->
    {noreply, State}.

%% @spec handle_cast(Msg, State) -> {noreply, State}
%%
handle_cast(stop, State) ->
    {stop, normal, State}.


%% @spec terminate(Reason, State) -> ok
%%
%% Close the socket for termination
%%
terminate(_Reason, State) ->
    #state{conn=C} = State,

    io:format("~p stopping~n", [?MODULE]),
    
    ok = pgsql:close(C).


%% @spec code_change(OldVersion, State, Extra) -> {ok, State}
%%
code_change(_OldVersion, State, _Extra) ->
    {ok, State}.

%% @spec stop(Name) -> {ok}
%%
stop(Name) ->
    gen_server:cast({global, Name}, stop).


log_message(C, Message) ->

    #message{message=Bin, parsed_values=Values} = Message,

    % Begin a transaction
    pg_begin(C),

    % Insert the NMEA sentence into the ais_raw table, setting the ais_seq sequence, used by later ais_msgX... tables
    {ok, RawInsertStmt} = pgsql:parse(C, "RawInsertStmt", "insert into ais_raw (id, nmea) values (nextval('ais_seq'), $1)", [text]),
    ok = pgsql:bind(C, RawInsertStmt, [binary_to_list(Bin)]),
    {ok, _} = pgsql:execute(C, RawInsertStmt),
    ok = pgsql:close(C, RawInsertStmt),
    ok = pgsql:sync(C),

    ok = insert_ais_message(C, Values),

    pg_commit(C),

    ok.


insert_ais_message(C, Values) ->
    Typemap = [{int, int4}, {float, float8}, {string, text}, {data, text}],

    [MessageTypeName] = [X || {messageTypeName, X} <- Values],

    [Fields] = [Fields || [{pattern, _}, {name, Mtn}, {fields, Fields}] <- ?AisPatterns, Mtn =:= MessageTypeName],
    FieldNames = [X||{X, _, _} <- Fields],

    Sql = io_lib:format("insert into ~p (~s) values (~s)", [
							  MessageTypeName,
							  string:join([atom_to_list(X) || X <- FieldNames], ", "),
							  string:join([io_lib:format("$~p", [X]) || X <- lists:seq(1, length(Fields))], ", ")
							]),
    %io:format("SQL: ~s\n", [Sql]),

    Types = lists:map(fun(Field) -> {_, FieldType, _} = Field, [Out] = [Value||{X, Value} <- Typemap, X =:= FieldType], Out end, Fields),

    %io:format("Types: ~p\n", [Types]),

    {ok, S} = pgsql:parse(C, "", Sql, Types),

    BindValues = lists:map(fun(FieldName) -> [Out] = [Value||{X, Value} <- Values, X =:= FieldName], Out end, FieldNames),

    %io:format("Values: ~p\n", [BindValues]),

    ok = pgsql:bind(C, S, BindValues),

    {ok, _} = pgsql:execute(C, S),
    ok = pgsql:close(C, S),
    ok = pgsql:sync(C),

    %io:format("\n", []),

    %io:format("Type: ~p\n", [MessageTypeName]),
    ok.


pg_begin(C) ->
    {ok, _, _} = pgsql:squery(C, "begin"),
    ok.

pg_commit(C) ->
    {ok, _, _} = pgsql:squery(C, "commit"),
    ok.
