%%%--------------------------------------------------------------------- 
%%% Copyright (C) 2008  Erik G. Burrows
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%
%%%--------------------------------------------------------------------- 
%%% Description module manager
%%%--------------------------------------------------------------------- 
%%% The manager is an erlang supervisor module that, through it's public
%%% API and file read/write operations starts, stops, connects and disconnects
%%% sensornet modules. 
%%%
%%%--------------------------------------------------------------------- 
%%% Exports
%%%--------------------------------------------------------------------- 
%%% stop() -> ok
%%%     Disconnect and shutdown all modules, then shutdown manager process
%%%
%%% start() -> ok
%%%     Start a manager process, wait for it. Returns on process exit.
%%%
%%% start(ConfigFile) -> ok
%%%     Start a manager process, specifying config file of
%%%     modules/subscriptions, wait for it. Returns on process exit.
%%%
%%% start_link() -> {ok, Pid}
%%%     Start a manager process, returns process ID.
%%%
%%% start_link(ConfigFile) -> {ok, Pid}
%%%     Start a manager process, specifying config file of
%%%     modules/subscriptions. Returns process ID.
%%%
%%% add(Name, Module, Args)
%%%     Load and start an I/O module.
%%%     Name = String name of module, used to specify subscriptions
%%%     Module = Atom name of module
%%%     Args = List of arguments to pass to module on startup. See module
%%%         documentation for required arguments.
%%%
%%% del(Name)
%%%     Remove any subscriptions referencing this module, and terminate it.
%%%
%%% connect(OutputName, OutputName, InputModule, InputName)
%%%     Subscribe the module named by InputModule to the module named by
%%%     OutputModule, connecting the output named OutputName to InputName.
%%%
%%% main()
%%%     Start a manager process, and parse command-line arguments for modules
%%%     to load, and subscriptions to setup.
%%%
%%%--------------------------------------------------------------------- 
%%% Revision History
%%%--------------------------------------------------------------------- 
%%% Revision 0.0.1 Date 2009/02/20 Author Erik Burrows erik@erikburrows.com
%%%     Initial revision, basic implementation of modular GNUradio-ish
%%%     structure.
%%%-------------------------------------------------------------------
%%%

-module(manager).
-behaviour(gen_server).


%% API
-export([
	 start_link/0,
	 start_link/1,
	 add/3,
	 del/1,
	 connect/4,
	 stop/0,
	 save/0,
	 save/1,
	 load/1,
	 main/0
	]).


%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
	 terminate/2, code_change/3]).


-define(SERVER, ?MODULE).

-record(state, {config_file, modules, subscriptions}).
-record(module, {name, type, args, pid}).
%-record(subscription, {output_module, output_name, input_module, input_name}).

%%====================================================================
%% API
%%====================================================================

main() ->

    Args = init:get_plain_arguments(),

    case Args of
	[] ->
	    io:format("Error: You must give me something to do!\n"),
	    usage(),
	    init:stop();
	_ ->
	    {ok, Pid} = start_link(),
	    parse_command_args(Args),

	    _ = receive
		{'EXIT', Pid, _} ->
		    {ok}
	    end,

	    init:stop()
    end.


add(Name, Type, Args) ->
    gen_server:call({global, ?SERVER}, {add, Name, Type, Args}).


del(Name) ->
    gen_server:call({global, ?SERVER}, {del, Name}).


connect(OutputModule, OutputName, InputModule, InputName) ->
    gen_server:call({global, ?SERVER}, {connect, OutputModule, OutputName, InputModule, InputName}).


start_link([ConfigFile]) ->
    gen_server:start_link({global, ?SERVER}, ?MODULE, [ConfigFile], []);

start_link(ConfigFile) ->
    gen_server:start_link({global, ?SERVER}, ?MODULE, [ConfigFile], []).

start_link() ->
    gen_server:start_link({global, ?SERVER}, ?MODULE, [], []).
 

stop() ->
    gen_server:cast({global, ?SERVER}, stop).


save() ->
    gen_server:call({global, ?SERVER}, {save}).


save(File) ->
    gen_server:call({global, ?SERVER}, {save, File}).


load(File) ->
    gen_server:call({global, ?SERVER}, {load, File}).


%%====================================================================
%% gen_server callbacks
%%====================================================================


init([]) ->
    process_flag(trap_exit, true),
    State = #state{modules=orddict:new(), subscriptions=ordsets:new()},
    {ok, State};

init([ConfigFile]) ->
    process_flag(trap_exit, true),
    State = #state{config_file=ConfigFile, modules=orddict:new(), subscriptions=ordsets:new()},

    case file:open(ConfigFile, read) of
	{ok, FD} ->
	    file:close(FD),
	    do_load(ConfigFile, State);

	{error, _} ->
	    {ok, State}
    end.


handle_call({save}, _From, State) ->
    {Reply, NewState} = do_save(State),
    {reply, Reply, NewState};

handle_call({save, File}, _From, State) ->
    {Reply, NewState} = do_save(File, State),
    {reply, Reply, NewState};

handle_call({load, File}, _From, State) ->
    {Reply, NewState} = do_load(File, State),
    {reply, Reply, NewState};

handle_call({add, Name, Type, Args}, _From, State) ->
    {Reply, NewState} = add_module(Name, Type, Args, State),
    {reply, Reply, NewState};

handle_call({del, Name}, _From, State) ->
    {Reply, NewState} = del_module(Name, State),
    {reply, Reply, NewState};

handle_call({connect, OutputModule, OutputName, InputModule, InputName}, _From, State) ->
    {Reply, NewState} = connect_module(OutputModule, OutputName, InputModule, InputName, State),
    {reply, Reply, NewState}.


handle_cast({'EXIT', Pid, Reason}, State) ->
    case restart_failed_module(Pid, Reason, State) of
	{ok, NewState} ->
	    {noreply, NewState};
	{error, NewState} ->
	    {noreply, NewState}
    end;

handle_cast({start_module, Name}, State) ->
	case start_module(Name, State) of
		{ok, NewState} ->
			{noreply, NewState};
		{error, NewState} ->
			{noreply, NewState}
	end;
    
handle_cast(stop, State) ->
    {stop, normal, State}.

% Retransmit the EXIT signal to ourself via a broadcast, so it's asynchronous
handle_info({'EXIT', Pid, Reason}, State) ->
    gen_server:cast({global, ?SERVER}, {'EXIT', Pid, Reason}),
    {noreply, State}.

terminate(_Reason, State) ->
    %io:format("Beginning system shutdown\n"),
    #state{modules=Modules} = State,
    ModuleList = [X || {X, _} <- orddict:to_list(Modules)],
    terminate_modules(ModuleList, State),
    %io:format("All modules shutdown.\n"),
    ok.


code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

do_save(State) ->
    #state{config_file=File} = State,

    case File of
	undefined ->
	    {{error, no_filename}, State};

	_ ->
	    do_save(File, State)
    end.

do_save(File, State) ->
    #state{modules=Modules, subscriptions=Subscriptions} = State,

    {ok, FD} = file:open(File, write),

    CleanModules = orddict:from_list(
		     [{Name, {V, W, X, Y, undefined}} || {Name, {V, W, X, Y, _}} <- orddict:to_list(Modules)]),

    ok = io:fwrite(FD, "~p.~n", [{{sensornet_conf_vers, 1}, {modules, CleanModules}, {subscriptions, Subscriptions}}]),

    ok = file:close(FD),
    
    {ok, State}.

do_load(File, State) ->
    {ok, [{{sensornet_conf_vers, 1}, {modules, Modules}, {subscriptions, Subscriptions}}]} = file:consult(File),
    {ok, NewState} = load_modules([X || {_, X} <- orddict:to_list(Modules)], State),
    {ok, NewerState} = load_subscriptions(ordsets:to_list(Subscriptions), NewState),
    {ok, NewerState}.


load_modules([], State) ->
    {ok, State};

load_modules([H|T], State) ->
    #module{name=Name, type=Type, args=Args} = H,
    {ok, NewState} = add_module(Name, Type, Args, State),
    load_modules(T, NewState).

load_subscriptions([], State) ->
    {ok, State};
load_subscriptions([H|T], State) ->
    {OutputModule, OutputName, InputModule, InputName} = H,
    {ok, NewState} = connect_module(OutputModule, OutputName, InputModule, InputName, State),
    load_subscriptions(T, NewState).


% add_module(Name, Type, Args, State) -> {ok | {error, Why}, State}
add_module(Name, Type, Args, State) ->
    #state{modules=Modules} = State,

    case orddict:find(Name, Modules) of

	% Return error if name is already in the modules list
	{ok, _} -> 
	    {{error, already_running}, State};

	% Name not found, ok to add.
	error ->
		Module = #module{name=Name, type=Type, args=Args},
		NewModules = orddict:store(Name, Module, Modules),
		NewState = State#state{modules=NewModules},
		gen_server:cast({global, ?SERVER}, {start_module, Name}),
		{ok, NewState}
    end.

start_module(Name, State) ->
    #state{modules=Modules, subscriptions=Subscriptions} = State,

    io:format("Starting module: ~p\n", [Name]),

    case orddict:find(Name, Modules) of
	{ok, Module} ->
		#module{type=Type, args=Args} = Module,

		case catch apply(Type, start_link, [Name | [Args]]) of
			{ok, Pid} ->
				NewModule = Module#module{pid=Pid},
				NewModules = orddict:store(Name, NewModule, Modules),
				NewState = State#state{modules=NewModules},
				subscribe([{W, X, Y, Z} || {W, X, Y, Z} <- ordsets:to_list(Subscriptions), W =:= Name], Modules),
				subscribe([{W, X, Y, Z} || {W, X, Y, Z} <- ordsets:to_list(Subscriptions), Y =:= Name], Modules),
				io:format("Module ~p started.\n", [Name]),
				{ok, NewState};
			Error ->
				io:format("Error starting module ~p:\n~p\n\n", [Name, Error]),
				timer:sleep(1000),
				gen_server:cast({global, ?SERVER}, {start_module, Name}),
				{ok, State}
		end;
	error ->
		{error, module_not_found}
    end.

% restart_failed_module(Pid, Reason, State) -> {ok | error, State}
restart_failed_module(Pid, _Reason, State) ->
    %io:format("restart_failed_module: Starting for pid ~p\n", [Pid]),

    #state{modules=Modules} = State,

    timer:sleep(1000),

    % Resolve the Name from the Pid
    case [Name || {Name, #module{pid=TmpPid}} <- orddict:to_list(Modules), TmpPid =:= Pid] of
	[] ->
	    {ok, State};

	[Name] -> 

	    io:format("Restarting module: ~p\n", [Name]),
    
            gen_server:cast({global, ?SERVER}, {start_module, Name}),
            {ok, State}
    end.

% del_module(Name, State) -> {ok | {error, Why}, State}
del_module(Name, State) ->
    #state{modules=Modules, subscriptions=Subscriptions} = State,

    case orddict:find(Name, Modules) of 
	{ok, Module} ->
	    #module{type=Type} = Module,

	    % Unsubscribe this module from any other modules
	    unsubscribe([{W, X, Y, Z} || {W, X, Y, Z} <- ordsets:to_list(Subscriptions), Y =:= Name], Modules),
	    unsubscribe([{W, X, Y, Z} || {W, X, Y, Z} <- ordsets:to_list(Subscriptions), W =:= Name], Modules),

	    NewSubscriptions = 
		ordsets:from_list([{W, X, Y, Z} || {W, X, Y, Z} <- ordsets:to_list(Subscriptions), W =/= Name, Y =/= Name]),

	    case catch apply(Type, stop, [Name]) of
		ok -> 

		    NewState = State#state{modules=orddict:erase(Name, Modules), subscriptions=NewSubscriptions},

		    {ok, NewState};

		Error ->
		    io:format("Error deleting module ~p:\n~p\n\n", [Name, Error]),
		    {{error, could_not_stop}, State}
	    end;

	{error} ->
	    {{error, no_such_module}, State}
    end.


subscribe([], _Modules) ->
    ok;

subscribe([Subscriber|T], Modules) ->
    {OutputModule, OutputName, InputModule, InputName} = Subscriber,
    io:format("Subscribing ~p:~p to ~p:~p\n", [OutputModule, OutputName, InputModule, InputName]),

    #module{type=Type} = orddict:fetch(OutputModule, Modules),
    _ = case catch apply(Type, subscribe, [OutputModule, OutputName, InputModule, InputName]) of
	ok ->
	    ok;
	Error ->
	    {dontcare, Error}
    end,

    subscribe(T, Modules).


unsubscribe([], _Modules) ->
    ok;

unsubscribe([Subscriber|T], Modules) ->
    {OutputModule, OutputName, InputModule, InputName} = Subscriber,
    %io:format("Unsubscribing ~p ~p ~p ~p\n", [OutputModule, OutputName, InputModule, InputName]),

    #module{type=Type} = orddict:fetch(OutputModule, Modules),
    _ = case catch apply(Type, unsubscribe, [OutputModule, OutputName, InputModule, InputName]) of
	ok ->
	    ok;
	Error ->
	    {dontcare, Error}
    end,

    unsubscribe(T, Modules).


% connect_module(OutputModule, OutputName, InputModule, InputName, State) -> {ok | {error, Why}, State}
connect_module(OutputModule, OutputName, InputModule, InputName, State) ->
    #state{modules=Modules, subscriptions=Subscriptions} = State,

    Id = {OutputModule, OutputName, InputModule, InputName},

    % FIXME: Check for type compatbility

    case orddict:find(OutputModule, Modules) of
	{ok, Module} ->

	    #module{type=Type} = Module,

	    case ordsets:is_element(Id, Subscriptions) of
		true ->
		    {{error, already_connected}, State};

		false ->

		    NewSubscriptions = ordsets:add_element(Id, Subscriptions),
		    NewState = State#state{subscriptions=NewSubscriptions},

		    io:format("Subscribing ~p:~p to ~p:~p\n", [OutputModule, OutputName, InputModule, InputName]),

		    case catch apply(Type, subscribe, [OutputModule, OutputName, InputModule, InputName]) of

			ok -> {ok, NewState};
			{'EXIT', {noproc, _}} -> {ok, NewState};
			Error ->
			    io:format("Error connecting ~p to ~p:\n~p\n\n", [OutputModule, InputModule, Error]),
			    {{error, connection_failed}, State}
		    end
	    end
    end.

% terminate_all_modules(State) -> {ok | error, State}
terminate_modules([], _State) ->
    ok;

terminate_modules([H|T], State) ->
    {ok, NewState} = del_module(H, State),
    terminate_modules(T, NewState).

usage() ->
    io:format("Usage: sensornet.sh [-load <filename>] [-module <type> <name> [arg1]:[arg2]...] [-connect <out module> <output name> <in module> <input name>] [-save <filename>]\n").


%%
%% @spec parse_command_args(Arglist) -> ok.
%%
%% Parses the comand-line, executing commands as found.
%%
parse_command_args([]) ->
    ok;

parse_command_args(["-load", Filename|T]) ->
    %io:format("Loading file: ~p\n", [Filename]),
    load(Filename),
    parse_command_args(T);

parse_command_args(["-help"|_]) ->
    usage(),
    init:stop();

parse_command_args(["--help"|_]) ->
    usage(),
    init:stop();

parse_command_args(["-?"|_]) ->
    usage(),
    init:stop();

parse_command_args(["-save", Filename|T]) ->
    %io:format("Saving to file: ~p\n", [Filename]),
    save(Filename),
    parse_command_args(T);

parse_command_args(["-connect", OutputModule, OutputName, InputModule, InputName|T]) ->
    %io:format("Connecting ~p:~p to ~p:~p\n", [OutputModule, OutputName, InputModule, InputName]),
    connect(OutputModule, OutputName, InputModule, InputName),
    parse_command_args(T);

parse_command_args(["-module", ModuleType, ModuleName, ModuleArgString|T]) ->
    %io:format("Loading module ~p as ~p\n", [ModuleType, ModuleName]),
    {ok, ModuleArgs} = regexp:split(ModuleArgString, ":"),
    add(ModuleName, list_to_atom(ModuleType), ModuleArgs),
    parse_command_args(T);

parse_command_args([H|_]) ->
    io:format("ERROR: Unknown option: ~p\n", [H]),
    usage(),
    init:stop().
