%%%--------------------------------------------------------------------- 
%%% Copyright (C) 2008  Erik G. Burrows
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%
%%%--------------------------------------------------------------------- 
%%% Description module nmea_ais_parser
%%%--------------------------------------------------------------------- 
%%% Filter module to pass through only specified sentence types.
%%%
%%%--------------------------------------------------------------------- 
%%% Exports
%%%--------------------------------------------------------------------- 
%%% @spec start_link(Name) -> {ok, PID}
%%%
%%%     gen_server wrapper for init() function. Instantiates the
%%%     module with the specified name, and passes in startup parameters
%%%
%%%     Name = Name of this instance of this module
%%%
%%% @spec input(Name, Message) -> {ok}
%%%     Submit a message for translation, will be translated and sent to subscribers
%%%     Name = Name of module instance
%%%     Message = Character array of message to send
%%%
%%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request a subscription to the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to subscribe to
%%%     Subscriber = Globally registered name of subscribing module
%%%     InputName = Name of input on subscribing module to send messages to
%%%
%%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request to unsubscribe from the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to unsubscribe from
%%%     Subscriber = Globally registered name of subscribed module
%%%     InputName = Name of input on subscribed module to unsubscribe
%%%
%%%--------------------------------------------------------------------- 
%%% Revision History
%%%--------------------------------------------------------------------- 
%%% Revision 0.0.1 Date 2009/02/20 Author Erik Burrows erik@erikburrows.com
%%%     Initial revision, basic implementation of modular GNUradio-ish
%%%     structure.
%%%
-module(nmea_ais_parser).


%% gen_server is used to stop/start module
-behavior(gen_server).


%% TCP is a bi-directional protocol, so this module is a source and a sink.
-behavior(gen_source).
-behavior(gen_sink).

-include_lib("ais_patterns.hrl").
-include_lib("common_records.hrl").



%% Public exported functions.
%%   input/3: Send a message to the remote host
%%   subscribe/4: Request input from this module to be sent to another
%%   unsubscribe/4: Remove a subscription
%%   init, stop, handle..: gen_server required public functions
%%
-export([
    start_link/2,
    input/5,
    subscribe/4,
    unsubscribe/4,

    to_int/1,
    to_bin/1,
    to_char/1,
    to_string/1,
    binary_to_bit_string/1,
    parse_sentence/2,

    init/1,
    stop/1,
    handle_call/3,
    handle_cast/2,
    handle_info/2,
    terminate/2,
    code_change/3
    ]).


%% Module input/output signatures, as required by manager program
%%
-inputs([{"NMEA", nmea_type}]).
-outputs([{"NMEA", nmea_type}]).


%% Data Type: state
%% Where:
%%   name = Name of this instance of this module (ex. Chartplotter)
%%   host = Remote host to connect to
%%   port = TCP port on remote host
%%   socket = Connected socket to remote host
%%   buf = Character buffer for receiver logic
%%
-record(state, {name, subscribers, buffer={0,0,""}}).


%% @spec input(Name, InputName, Message) -> {ok}
%%
%% Send a message to the module, to be sent to the remote host.
%%
%% Name = Name of this instance of this module
%% InputName = Name of the input to send to
%%
input(Name, InputName, OutputModule, OutputName, Message) -> 
    gen_server:call({global, Name}, {input, InputName, OutputModule, OutputName, Message}).


%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request a subscription to the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to subscribe to
%% Subscriber = Globally registered name of subscribing module
%% InputName = Name of input on subscribing module to send messages to
%%
subscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {subscribe, OutputName, Subscriber, InputName}).


%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request to unsubscribe from the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to unsubscribe from
%% Subscriber = Globally registered name of subscribed module
%% InputName = Name of input on subscribed module to unsubscribe
%%
unsubscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {unsubscribe, OutputName, Subscriber, InputName}).


%% @spec start_link(Name, [Sentences]) -> {ok, PID}
%%
%% gen_server wrapper for init() function. Instantiates the
%% module with the specified name, and passes in startup parameters
%%
%% Sentences = NMEA Sentence prefixes to let through
%%
start_link(Name, Sentences) ->
    gen_server:start_link({global, Name}, ?MODULE, [Name | [Sentences]], []).


%% @spec init([Name, Sentences]) -> {ok, State}
%%
%% Startup function, called only by start_link RPC wrapper.
%%
init([Name, _]) ->
    io:format("~p starting~n", [?MODULE]),

    Outputs = [Y || {Y, _} <- lists:concat([X || {outputs, X} <- module_info(attributes)])],
    Sublist = subscriptions:new(Outputs),

    {ok, #state{name=Name, subscribers=Sublist}}.


%% @spec handle_call({name, Arg, ...}) -> {reply, ok, Arg, ...}
%%
%% Accept message to input named NMEA
%%
handle_call({input, "NMEA", _OutputModule, _OutputName, Message}, _From, State) ->
    #state{subscribers=Subscribers, name=Name} = State,

    %io:format("~p: Received message: ~p\n", [?MODULE, Message#message.message]),

    % Append this module's name to the message's internal filters list (path through the system)
    #message{filtering_modules=FilterList} = Message,
    NewFilterList = FilterList ++ [?MODULE],

    NewState = case parse_sentence(State, binary_to_list(Message#message.message)) of
	error ->
	    % Parsing error, nothing to do, but not an error to pass up.
	    State;
	{noop, NewStateVar} ->
	    % Some messages require more than one packet, so this is not an error.
	    NewStateVar;
	{ok, NewStateVar, Values} ->
	    subscriptions:send([Message#message{filtering_modules=NewFilterList, parsed_values=Values}], Name, "NMEA", Subscribers),
            NewStateVar
    end,

    {reply, ok, NewState};


%% Handle subscription requests
%%
handle_call({subscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:subscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}};


%% Handle unsubscribe requests
%%
handle_call({unsubscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:unsubscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}}.


%% @spec handle_cast(Msg, State) -> {noreply, State}
%%
handle_cast(stop, State) ->
    {stop, normal, State}.


%% @spec terminate(Reason, State) -> ok
%%
%% Close the socket for termination
%%
terminate(_Reason, _State) ->
    io:format("~p stopping~n", [?MODULE]),
    ok.


%% @spec code_change(OldVersion, State, Extra) -> {ok, State}
%%
code_change(_OldVersion, State, _Extra) ->
    {ok, State}.

%% @spec stop(Name) -> {ok}
%%
stop(Name) ->
    gen_server:cast({global, Name}, stop).


handle_info(_Info, State) ->
    {noreply, State}.


%%
%% Internal functions
%%

decode_pattern(State, _BitString, Values, []) ->
    {ok, State, Values};

% decode integer
decode_pattern(State, BitString, Values, [{Name, int, [{min_length, MinLength},{max_length, _MaxLength}]}|T]) ->
    if
        length(BitString) >= MinLength ->
            Val = to_int(string:substr(BitString, 1, MinLength)),
            %io:format("Decoded ~p as ~p from ~p\n", [Name, Val, string:substr(BitString, 1, MinLength)]),
            decode_pattern(State, string:substr(BitString, MinLength+1), [{Name, Val} | Values], T);
        true ->
            error
     end;

% decode float
decode_pattern(State, BitString, Values, [{Name, float, [{min_length, MinLength}, {max_length, _MaxLength}, {divisor, Divisor}]}|T]) ->
    if
        length(BitString) >= MinLength ->
            Val = to_int(string:substr(BitString, 1, MinLength)) - case string:substr(BitString, 1, 1) of
                "0" ->
                    0;
                "1" ->
                    math:pow(2, MinLength)
            end,

            %io:format("Decoded ~p as ~p from ~p\n", [Name, Val / Divisor, string:substr(BitString, 1, MinLength)]),
            decode_pattern(State, string:substr(BitString, MinLength+1), [{Name, Val / Divisor} | Values], T);
        true ->
	    error
    end;
    

% decode string
decode_pattern(State, BitString, Values, [{Name, string, [{min_length, MinLength}, {max_length, _MaxLength}]}|T]) ->
    if
        length(BitString) >= MinLength ->
            Val = to_string(string:substr(BitString, 1, MinLength)),    
            %io:format("Decoded ~p as ~p from ~p\n", [Name, Val, string:substr(BitString, 1, MinLength)]),
            decode_pattern(State, string:substr(BitString, MinLength+1), [{Name, Val} | Values], T);
        true ->
	    error
    end;

% decode data
decode_pattern(State, BitString, Values, [{Name, data, [{min_length, MinLength}, {max_length, MaxLength}]}|T]) ->
    if
        length(BitString) >= MinLength ->
	    if
		MaxLength > MinLength ->
		    Val = BitString,
		    decode_pattern(State, string:substr(BitString, MinLength+1), [{Name, Val} | Values], T);
		true ->
		    Val = string:substr(BitString, 1, MinLength),
		    %io:format("Decoded ~p as ~p from ~p\n", [Name, Val, string:substr(BitString, 1, MinLength)]),
		    decode_pattern(State, string:substr(BitString, MinLength+1), [{Name, Val} | Values], T)
	    end;
        true ->
	    error
    end.


parse_sentence(State, NmeaAis) ->
    case regexp:match(NmeaAis, "^\!AIVDM") of
        {match, _,_} ->

	    case regexp:split(NmeaAis, ",") of
		{ok, [_, MessageMaxStr, MessageNumStr, _, _, Payload, Checksum]} ->

		    {MessageMax, _} = string:to_integer(MessageMaxStr),
		    {MessageNum, _} = string:to_integer(MessageNumStr),

		    if 
			MessageNum == MessageMax ->
			    
			    #state{buffer={OldMessageMax, OldMessageNum, OldPayload}} = State,

  		            NewPayload = if
					     (MessageNum > 1) 
					     and (OldMessageMax == MessageMax) 
					     and (OldMessageNum == MessageNum -1) ->

						 string:concat(OldPayload, Payload);

					     true ->
						 Payload
					 end,

                            Bitstring = binary_to_bit_string(NewPayload),
			    
                            parse_sentence_pattern(State, Bitstring, ?AisPatterns);

                        MessageNum < MessageMax ->

			    #state{buffer={OldMessageMax, OldMessageNum, OldPayload}} = State,

			    NewState = if
					   (MessageNum > 1)
					   and (OldMessageMax == MessageMax)
					   and (OldMessageNum == MessageNum -1) ->
					       State#state{buffer={MessageMax, MessageNum, string:concat(OldPayload, Payload)}};

					   true ->
					       State#state{buffer={MessageMax, MessageNum, Payload}}
				       end,

                            {noop, NewState}
                    end;

		_Else ->
		    error
            end;

        _Else ->
            error
    end.


parse_sentence_pattern(State, _Bitstring, []) ->
    {ok, State, ""};

parse_sentence_pattern(State, Bitstring, [H|T]) ->
    [{pattern, Pattern}, {name, Name}, {fields, Fields}] = H,
    case regexp:match(Bitstring, Pattern) of
        {match, _, _} ->
            decode_pattern(State, Bitstring, [{messageTypeName, Name}], Fields);
        nomatch ->
            parse_sentence_pattern(State, Bitstring, T);
        error ->
            error
    end.


binary_to_bit_string([]) ->
    [];

binary_to_bit_string([H|T]) ->

    SixBit = if
        (H >= 48) and (H < 88) ->
             H - 48;
        (H > 96) and (H < 121) ->
             H - 56;
        true ->
             0
    end,

    Binary = to_bin(SixBit),

    Padded = string:right(Binary, 6, $0),

    string:concat(Padded, binary_to_bit_string(T)).


to_bin(Int) ->
    to_bin(Int, []).

to_bin(0, Buff) -> Buff;
to_bin(Int, Buff) ->
    to_bin(Int div 2, [to_char(Int rem 2)|Buff]).

to_char(Int) when Int > 9 -> Int rem 10 + $A;
to_char(Int)              -> Int + $0.


to_int(Binary) ->
   to_int(0, Binary).


to_int(Sum, "") ->
   Sum;

to_int(Sum, Binary) ->
   In = string:substr(Binary, 1, 1),
   {Val, _} = string:to_integer(In),
   to_int(Sum + (trunc(Val * math:pow(2, length(Binary)-1))), string:substr(Binary, 2)).

%
% i2b(Int, Length) -> String
%
% Convert an integer into a string of binary digits padded to Length size
%
i2b(Int, Length) -> string:right(to_bin(Int), Length, $0).


%
% to_string(Binary) -> String
%
% Convert a string of binary digits into a character string using six-bit encoding
%
to_string(Binary) ->
    to_string("", Binary).

to_string(Buf, Binary) when length(Binary) < 6 ->
    Buf;

to_string(Buf, Binary) ->

    Int = to_int(string:substr(Binary, 1, 6)),

    Char = if
        (Int < 32) and (Int >= 0) ->
            Int + 64;
        true ->
            Int
    end,

    NewChar = case Char of
        64 ->
          "";
        _Default ->
          string:chars(Char, 1)
    end,

    to_string(string:concat(Buf, NewChar), string:substr(Binary, 7)).

