%%%--------------------------------------------------------------------- 
%%% Copyright (C) 2008  Erik G. Burrows
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%
%%%--------------------------------------------------------------------- 
%%% Description module tcp_nmea_client
%%%--------------------------------------------------------------------- 
%%% Filter module to pass through only specified sentence types.
%%%
%%%--------------------------------------------------------------------- 
%%% Exports
%%%--------------------------------------------------------------------- 
%%% @spec start_link(Name, [Sentences]) -> {ok, PID}
%%%
%%%     gen_server wrapper for init() function. Instantiates the
%%%     module with the specified name, and passes in startup parameters
%%%
%%%     Name = Name of this instance of this module
%%%     Sentences = List of string prefixes to allow through
%%%
%%% @spec input(Name, Message) -> {ok}
%%%     Send a message to the remote server.
%%%     Name = Name of module instance
%%%     Message = Character array of message to send
%%%
%%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request a subscription to the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to subscribe to
%%%     Subscriber = Globally registered name of subscribing module
%%%     InputName = Name of input on subscribing module to send messages to
%%%
%%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request to unsubscribe from the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to unsubscribe from
%%%     Subscriber = Globally registered name of subscribed module
%%%     InputName = Name of input on subscribed module to unsubscribe
%%%
%%%--------------------------------------------------------------------- 
%%% Revision History
%%%--------------------------------------------------------------------- 
%%% Revision 0.0.1 Date 2009/02/20 Author Erik Burrows erik@erikburrows.com
%%%     Initial revision, basic implementation of modular GNUradio-ish
%%%     structure.
%%%
-module(nmea_filter).


%% gen_server is used to stop/start module
-behavior(gen_server).


%% TCP is a bi-directional protocol, so this module is a source and a sink.
-behavior(gen_source).
-behavior(gen_sink).


-include_lib("common_records.hrl").



%% Public exported functions.
%%   input/3: Send a message to the remote host
%%   subscribe/4: Request input from this module to be sent to another
%%   unsubscribe/4: Remove a subscription
%%   init, stop, handle..: gen_server required public functions
%%
-export([
    start_link/2,
    input/5,
    subscribe/4,
    unsubscribe/4,

    init/1,
    stop/1,
    handle_call/3,
    handle_cast/2,
    handle_info/2,
    terminate/2,
    code_change/3
    ]).


%% Module input/output signatures, as required by manager program
%%
-inputs([{"NMEA", nmea_type}]).
-outputs([{"NMEA", nmea_type}]).


%% Data Type: state
%% Where:
%%   name = Name of this instance of this module (ex. Chartplotter)
%%   host = Remote host to connect to
%%   port = TCP port on remote host
%%   socket = Connected socket to remote host
%%   buf = Character buffer for receiver logic
%%
-record(state, {name, sentences, subscribers}).


%% @spec input(Name, InputName, Message) -> {ok}
%%
%% Send a message to the module, to be sent to the remote host.
%%
%% Name = Name of this instance of this module
%% InputName = Name of the input to send to
%%
input(Name, InputName, OutputModule, OutputName, Message) -> 
    gen_server:call({global, Name}, {input, InputName, OutputModule, OutputName, Message}).


%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request a subscription to the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to subscribe to
%% Subscriber = Globally registered name of subscribing module
%% InputName = Name of input on subscribing module to send messages to
%%
subscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {subscribe, OutputName, Subscriber, InputName}).


%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request to unsubscribe from the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to unsubscribe from
%% Subscriber = Globally registered name of subscribed module
%% InputName = Name of input on subscribed module to unsubscribe
%%
unsubscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {unsubscribe, OutputName, Subscriber, InputName}).


%% @spec start_link(Name, [Sentences]) -> {ok, PID}
%%
%% gen_server wrapper for init() function. Instantiates the
%% module with the specified name, and passes in startup parameters
%%
%% Sentences = NMEA Sentence prefixes to let through
%%
start_link(Name, Sentences) ->
    gen_server:start_link({global, Name}, ?MODULE, [Name | [Sentences]], []).


%% @spec init([Name, Sentences]) -> {ok, State}
%%
%% Startup function, called only by start_link RPC wrapper.
%%
init([Name, Sentences]) ->
    io:format("~p starting (~p)~n", [?MODULE, Sentences]),

    Outputs = [Y || {Y, _} <- lists:concat([X || {outputs, X} <- module_info(attributes)])],
    Sublist = subscriptions:new(Outputs),

    {ok, #state{name=Name, sentences=Sentences, subscribers=Sublist}}.


%% @spec handle_call({name, Arg, ...}) -> {reply, ok, Arg, ...}
%%
%% Accept message to input named NMEA
%%
handle_call({input, "NMEA", _OutputModule, _OutputName, Message}, _From, State) ->
    #state{sentences=Sentences, subscribers=Subscribers, name=Name} = State,

    #message{filtering_modules=FilterList} = Message,

    NewFilterList = FilterList ++ [?MODULE],

    case check_sentence(Sentences, Message) of
	true ->
	    subscriptions:send([Message#message{filtering_modules=NewFilterList}], Name, "NMEA", Subscribers);
	false ->
	    ok
    end,

    {reply, ok, State};


%% Handle subscription requests
%%
handle_call({subscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:subscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}};


%% Handle unsubscribe requests
%%
handle_call({unsubscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:unsubscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}}.


%% @spec handle_cast(Msg, State) -> {noreply, State}
%%
handle_cast(stop, State) ->
    {stop, normal, State}.


%% @spec terminate(Reason, State) -> ok
%%
%% Close the socket for termination
%%
terminate(_Reason, _State) ->
    io:format("~p stopping~n", [?MODULE]),
    ok.


%% @spec code_change(OldVersion, State, Extra) -> {ok, State}
%%
code_change(_OldVersion, State, _Extra) ->
    {ok, State}.

%% @spec stop(Name) -> {ok}
%%
stop(Name) ->
    gen_server:cast({global, Name}, stop).


handle_info(_Info, State) ->
    {noreply, State}.


%%
%% Internal functions
%%
check_sentence([], _Message) ->
    false;

check_sentence([H|T], Message) ->
    #message{message=Bin} = Message,
    case lists:sublist(binary_to_list(Bin), 2, length(H)) =:= H of
	true ->
	    true;
	false ->
	    check_sentence(T, Message)
    end.
