%%%--------------------------------------------------------------------- 
%%% Copyright (C) 2008  Erik G. Burrows
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%
%%%--------------------------------------------------------------------- 
%%% Description module printer
%%%--------------------------------------------------------------------- 
%%% Logfile module, allows timestamped, input-module stamped message-by-message
%%% logging to a file.
%%%
%%%--------------------------------------------------------------------- 
%%% Exports
%%%--------------------------------------------------------------------- 
%%% @spec start_link(Name) -> {ok, PID}
%%%
%%%     gen_server wrapper for init() function. Instantiates the
%%%     module with the specified name, and passes in startup parameters
%%%
%%%     Name = Name of this instance of this module
%%%
%%% @spec input(Name, Message) -> {ok}
%%%     Send a message to the remote server.
%%%     Name = Name of module instance
%%%     Message = Character array of message to send
%%%
%%%--------------------------------------------------------------------- 
%%% Revision History
%%%--------------------------------------------------------------------- 
%%% Revision 0.0.1 Date 2009/08/01 Author Erik Burrows erik@erikburrows.com
%%%     Initial revision, basic implementation of modular GNUradio-ish
%%%     structure.
%%%
-module(printer).


%% gen_server is used to stop/start module
-behavior(gen_server).


%% This module sends but does not receive UDP packets, so it's a sink
-behavior(gen_sink).


-include_lib("common_records.hrl").



%% Public exported functions.
%%   input/3: Send a message to the remote host
%%   init, stop, handle..: gen_server required public functions
%%
-export([
    start_link/2,
    input/5,

    init/1,
    stop/1,
    handle_call/3,
    handle_cast/2,
    handle_info/2,
    terminate/2,
    code_change/3
    ]).


%% Module input/output signatures, as required by manager program
%%
-inputs([{"NMEA", nmea_type}]).
-outputs([]).


%% Data Type: state
%% Where:
%%   name = Name of this instance of this module (ex. Chartplotter)
%%   host = Remote host to connect to
%%   port = TCP port on remote host
%%   socket = Opened client socket
%%
-record(state, {name, prefix}).


%% @spec input(Name, InputName, Message) -> {ok}
%%
%% Send a message to the module, to be sent to the remote host.
%%
%% Name = Name of this instance of this module
%% InputName = Name of the input to send to
%%
input(Name, InputName, OutputModule, OutputName, Message) -> 
    gen_server:call({global, Name}, {input, InputName, OutputModule, OutputName, Message}).


%% @spec start_link(Name, [Prefix]) -> {ok, PID}
%%
%% gen_server wrapper for init() function. Instantiates the
%% module with the specified name, and passes in startup parameters
%%
start_link(Name, [Prefix]) ->
    gen_server:start_link({global, Name}, ?MODULE, [Name | [[Prefix]]], []).


%% @spec init([Name, []]) -> {ok, State}
%%
%% Startup function, called only by start_link RPC wrapper.
%%
init([Name, [Prefix]]) ->
    io:format("~p starting~n", [?MODULE]),

    {ok, #state{name=Name, prefix=Prefix}}.


%% @spec handle_call({name, Arg, ...}) -> {reply, ok, Arg, ...}
%%
%% Accept message to input named NMEA
%%
handle_call({input, "NMEA", _OutputModule, _OutputName, Message}, _From, State) ->
    #state{prefix=Prefix} = State,
    #message{originating_module=OriginatingModule, message=Bin, parsed_values=Values} = Message,

    {{Year, Month, Day}, {Hours, Minutes, Seconds}} = calendar:now_to_local_time(now()),

    io:format("~p: ~4.1.0s/~2.1.0s/~2.1.0s ~2.1.0s:~2.1.0s:~2.1.0s ~p ~9999p: ~p\n", [Prefix, io_lib:format("~p", [Year]), io_lib:format("~p", [Month]), io_lib:format("~p", [Day]), io_lib:format("~p", [Hours]), io_lib:format("~p", [Minutes]), io_lib:format("~p", [Seconds]), OriginatingModule, Values, Bin]),

    {reply, ok, State}.

% @spec handle_info(Info, State) -> {noreply, State}
%
handle_info(_Info, State) ->
    {noreply, State}.

%% @spec handle_cast(Msg, State) -> {noreply, State}
%%
handle_cast(stop, State) ->
    {stop, normal, State}.


%% @spec terminate(Reason, State) -> ok
%%
%% Close the socket for termination
%%
terminate(_Reason, _State) ->
    io:format("~p stopping~n", [?MODULE]).

%% @spec code_change(OldVersion, State, Extra) -> {ok, State}
%%
code_change(_OldVersion, State, _Extra) ->
    {ok, State}.

%% @spec stop(Name) -> {ok}
%%
stop(Name) ->
    gen_server:cast({global, Name}, stop).
