%%%--------------------------------------------------------------------- 
%%% Copyright (C) 2008  Erik G. Burrows
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%
%%%--------------------------------------------------------------------- 
%%% Description module tcp_nmea_client
%%%--------------------------------------------------------------------- 
%%% TCP NMEA server module, implements source and sink functions.
%%% When started, the module opens a TCP socket to accept connections on
%%% the port specified. Any connections are sent all messages from the "NMEA"
%%% input, and any NMEA sentences from any client are sent to the "NMEA"
%%% output.
%%%
%%%
%%%--------------------------------------------------------------------- 
%%% Exports
%%%--------------------------------------------------------------------- 
%%% @spec start_link(Name, [Port]) -> {ok, PID}
%%%
%%%     gen_server wrapper for init() function. Instantiates the
%%%     module with the specified name, and passes in startup parameters
%%%
%%%     Name = Name of this instance of this module
%%%     Port = TCP port number or name on which to listen
%%%       * Specified in string format
%%%
%%% @spec input(Name, Message) -> {ok}
%%%     Send a message to the remote server.
%%%     Name = Name of module instance
%%%     Message = Character array of message to send
%%%
%%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request a subscription to the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to subscribe to
%%%     Subscriber = Globally registered name of subscribing module
%%%     InputName = Name of input on subscribing module to send messages to
%%%
%%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request to unsubscribe from the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to unsubscribe from
%%%     Subscriber = Globally registered name of subscribed module
%%%     InputName = Name of input on subscribed module to unsubscribe
%%%
%%%--------------------------------------------------------------------- 
%%% Revision History
%%%--------------------------------------------------------------------- 
%%% Revision 0.0.1 Date 2009/02/20 Author Erik Burrows erik@erikburrows.com
%%%     Initial revision, basic implementation of modular GNUradio-ish
%%%     structure.
%%%
-module(tcp_nmea_server).


%% gen_server is used to stop/start module
-behavior(gen_server).


%% TCP is a bi-directional protocol, so this module is a source and a sink.
-behavior(gen_source).
-behavior(gen_sink).

-include_lib("common_records.hrl").


%% Public exported functions.
%%   input/3: Send a message to the remote host
%%   subscribe/4: Request input from this module to be sent to another
%%   unsubscribe/4: Remove a subscription
%%   init, stop, handle..: gen_server required public functions
%%
-export([
    start_link/2,
    input/5,
    subscribe/4,
    unsubscribe/4,

    init/1,
    stop/1,
    handle_call/3,
    handle_cast/2,
    handle_info/2,
    terminate/2,
    code_change/3
    ]).


%% Module input/output signatures, as required by manager program
%%
-inputs([{"NMEA", nmea_type}]).
-outputs([{"NMEA", nmea_type}]).


%% Data Type: state
%% Where:
%%   name = Name of this instance of this module
%%   port = TCP port to listen on
%%   socket = Listen socket
%%   client_sockets = ordered dictionary of client sockets and input buffers
%%
-record(state, {name, port, socket, subscribers, client_sockets}).


%% @spec input(Name, InputName, Message) -> {ok}
%%
%% Send a message to the module, to be sent to all connected clients.
%%
%% Name = Name of this instance of this module
%% InputName = Name of the input to send to
%%
input(Name, InputName, OutputModule, OutputName, Message) -> 
    gen_server:call({global, Name}, {input, InputName, OutputModule, OutputName, Message}).


%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request a subscription to the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to subscribe to
%% Subscriber = Globally registered name of subscribing module
%% InputName = Name of input on subscribing module to send messages to
%%
subscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {subscribe, OutputName, Subscriber, InputName}).


%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request to unsubscribe from the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to unsubscribe from
%% Subscriber = Globally registered name of subscribed module
%% InputName = Name of input on subscribed module to unsubscribe
%%
unsubscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {unsubscribe, OutputName, Subscriber, InputName}).


%% @spec start_link(Name, [Port]) -> {ok, PID}
%%
%% gen_server wrapper for init() function. Instantiates the
%% module with the specified name, and passes in startup parameters
%%
%% Name = Name of this instance of this module
%%
start_link(Name, [Port]) ->
    gen_server:start_link({global, Name}, ?MODULE, [Name | [[Port]]], []).


%% @spec init([Name, [Port]]) -> {ok, State}
%%
%% Startup function, called only by start_link RPC wrapper.
%%
init([Name, [Port]]) ->
    io:format("~p starting~n", [?MODULE]),

    {ok, Socket} = gen_tcp:listen(list_to_integer(Port), [binary, {active, true}, {packet, 0}]),

    MyPID = self(),
    spawn(fun() -> accept_connections(Socket, MyPID) end),

    Outputs = [Y || {Y, _} <- lists:concat([X || {outputs, X} <- module_info(attributes)])],
    Sublist = subscriptions:new(Outputs),

    ClientSockets = orddict:new(),

    {ok, #state{name=Name, socket=Socket, subscribers=Sublist, client_sockets=ClientSockets}}.


%% @spec handle_call({name, Arg, ...}) -> {reply, ok, Arg, ...}
%%
%% Accept message to input named NMEA
%%
handle_call({input, "NMEA", _OutputModule, _OutputName, Message}, _From, State) ->
    #state{client_sockets=Clients} = State,

    ClientSockets = [X || {X, _} <- orddict:to_list(Clients)],

    send_to_sockets(ordsets:to_list(ClientSockets), Message),

    {reply, ok, State};


%% Handle subscription requests
%%
handle_call({subscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:subscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}};


%% Handle unsubscribe requests
%%
handle_call({unsubscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:unsubscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}}.


%% @spec handle_info(Info, State) -> {noreply, State}
%%
%% Receive tcp message data from the a remote host. Accumulate data
%% in state "buf" element. Look for NMEA end-of-line chars (CRLF).
%% If found, send sentence to subscribers.
%%
handle_info({tcp, Socket, Bin}, State) ->
    #state{subscribers=Subscribers, client_sockets=Clients, name=MyName} = State,

    InBuf = orddict:fetch(Socket, Clients),

    Buf = [InBuf|binary_to_list(Bin)],

    % Look for a sentence
    {Sentences, Rest} = nmea:extract_sentences(Buf),

    subscriptions:send(pack_messages(MyName, Sentences), MyName, "NMEA", Subscribers),

    NewClients = orddict:store(Socket, Rest, Clients),

    {noreply, State#state{client_sockets=NewClients}};

handle_info({new_socket, Socket}, State) ->
    #state{client_sockets=ClientSockets} = State,
    NewClientSockets = orddict:store(Socket, [], ClientSockets),
    {noreply, State#state{client_sockets=NewClientSockets}};

handle_info({tcp_closed, Socket}, State) ->
    #state{client_sockets=ClientSockets} = State,
    NewClientSockets = orddict:erase(Socket, ClientSockets),
    {noreply, State#state{client_sockets=NewClientSockets}}.


%% @spec handle_cast(Msg, State) -> {noreply, State}
%%
handle_cast(stop, State) ->
    {stop, normal, State}.


%% @spec terminate(Reason, State) -> ok
%%
%% Close the socket for termination
%%
terminate(_Reason, State) ->
    io:format("~p stopping~n", [?MODULE]),
    #state{socket=Socket} = State,
    ok = gen_tcp:close(Socket).


%% @spec code_change(OldVersion, State, Extra) -> {ok, State}
%%
code_change(_OldVersion, State, _Extra) ->
    {ok, State}.

%% @spec stop(Name) -> {ok}
%%
stop(Name) ->
    gen_server:cast({global, Name}, stop).


%%
%% Internal functions
%%

accept_connections(Socket, Parent) ->
    case gen_tcp:accept(Socket) of
	{error, closed} ->
	    {error, closed};

	{ok, Sock} ->
	    gen_tcp:controlling_process(Sock, Parent),
	    Parent ! {new_socket, Sock},
	    accept_connections(Socket, Parent)
    end.


send_to_sockets([], _Message) ->
    ok;

send_to_sockets([Socket|Sockets], Message) ->
    #message{message=Bin} = Message,
    
    gen_tcp:send(Socket, Bin),
    send_to_sockets(Sockets, Message).


pack_messages(_, []) ->
    [];

pack_messages(Name, [H|T]) ->
    Message = #message{
      originating_module=Name,
      type=nmea,
      message=H
     },
    
    [Message] ++ pack_messages(Name, T).
