%%%--------------------------------------------------------------------- 
%%% Copyright (C) 2008  Erik G. Burrows
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%
%%%--------------------------------------------------------------------- 
%%% Description module tcp_nmea_client
%%%--------------------------------------------------------------------- 
%%% TCP NMEA Client module, implements source and sink functions.
%%% When started, the module makes a TCP connection to the specified
%%% host/port, and sends an optional initialization string. All received
%%% NMEA sentences are published to subscribers. 
%%%
%%% Messages received through input/2 are sent out to the server as-is.
%%%
%%%--------------------------------------------------------------------- 
%%% Exports
%%%--------------------------------------------------------------------- 
%%% @spec start_link(Name, [Port]) -> {ok, PID}
%%%
%%%     gen_server wrapper for init() function. Instantiates the
%%%     module with the specified name, and passes in startup parameters
%%%
%%%     Name = Name of this instance of this module
%%%     Port = Port name/number to on which to listen for packets
%%%       * Port number should be in string format
%%%
%%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request a subscription to the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to subscribe to
%%%     Subscriber = Globally registered name of subscribing module
%%%     InputName = Name of input on subscribing module to send messages to
%%%
%%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%%
%%%     Request to unsubscribe from the named output
%%% 
%%%     Name = Name of this instance of this module
%%%     OutputName = Name of the output of this module to unsubscribe from
%%%     Subscriber = Globally registered name of subscribed module
%%%     InputName = Name of input on subscribed module to unsubscribe
%%%
%%%--------------------------------------------------------------------- 
%%% Revision History
%%%--------------------------------------------------------------------- 
%%% Revision 0.0.1 Date 2009/02/20 Author Erik Burrows erik@erikburrows.com
%%%     Initial revision, basic implementation of modular GNUradio-ish
%%%     structure.
%%%
-module(udp_nmea_server).


%% gen_server is used to stop/start module
-behavior(gen_server).


%% TCP is a bi-directional protocol, so this module is a source and a sink.
-behavior(gen_source).

-include_lib("common_records.hrl").


%% Public exported functions.
%%   subscribe/4: Request input from this module to be sent to another
%%   unsubscribe/4: Remove a subscription
%%   init, stop, handle..: gen_server required public functions
%%
-export([
    start_link/2,
    subscribe/4,
    unsubscribe/4,

    init/1,
    stop/1,
    handle_call/3,
    handle_cast/2,
    handle_info/2,
    terminate/2,
    code_change/3
    ]).


%% Module input/output signatures, as required by manager program
%%
-inputs([]).
-outputs([{"NMEA", nmea_type}]).


%% Data Type: state
%% Where:
%%   name = Name of this instance of this module (ex. Chartplotter)
%%   host = Remote host to connect to
%%   port = TCP port on remote host
%%   socket = Connected socket to remote host
%%
-record(state, {name, host, port, socket, subscribers}).


%% @spec subscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request a subscription to the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to subscribe to
%% Subscriber = Globally registered name of subscribing module
%% InputName = Name of input on subscribing module to send messages to
%%
subscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {subscribe, OutputName, Subscriber, InputName}).


%% @spec unsubscribe(Name, OutputName, Subscriber, InputName) -> {ok}
%%
%% Request to unsubscribe from the named output
%% 
%% Name = Name of this instance of this module
%% OutputName = Name of the output of this module to unsubscribe from
%% Subscriber = Globally registered name of subscribed module
%% InputName = Name of input on subscribed module to unsubscribe
%%
unsubscribe(Name, OutputName, Subscriber, InputName) ->
    gen_server:call({global, Name}, {unsubscribe, OutputName, Subscriber, InputName}).


%% @spec start_link(Name, [Port]) -> {ok, PID}
%%
%% gen_server wrapper for init() function. Instantiates the
%% module with the specified name, and passes in startup parameters
%%
%% Name = Name of this instance of this module
%% Port = Remote host TCP port (name or number)
%%
start_link(Name, [Port]) ->
    gen_server:start_link({global, Name}, ?MODULE, [Name | [[Port]]], []).


%% @spec init([Name, Port]) -> {ok, State}
%%
%% Startup function, called only by start_link RPC wrapper.
%%
init([Name, [Port]]) ->
    io:format("~p starting~n", [?MODULE]),

    {ok, Socket} = gen_udp:open(list_to_integer(Port), [binary, {active, true}]),

    Outputs = [Y || {Y, _} <- lists:concat([X || {outputs, X} <- module_info(attributes)])],
    Sublist = subscriptions:new(Outputs),

    {ok, #state{name=Name, port=Port, socket=Socket, subscribers=Sublist}}.


%% Handle subscription requests
%%
handle_call({subscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:subscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}};


%% Handle unsubscribe requests
%%
handle_call({unsubscribe, OutputName, Subscriber, InputName}, _From, State) ->
    #state{subscribers=Subscribers} = State,
    NewSubscribers = subscriptions:unsubscribe({Subscriber, InputName}, OutputName, Subscribers),
    {reply, ok, State#state{subscribers=NewSubscribers}}.


%% @spec handle_info(Info, State) -> {noreply, State}
%%
%% Receive udp message data from the remote host. Notify subscribers.
%%
handle_info({udp, _Socket, _Ip, _InPortNo, Bin}, State) ->
    #state{name=Name, subscribers=Subscribers} = State,

    Message = #message{
      originating_module=Name,
      type=nmea,
      message=Bin
     },

    subscriptions:send([Message], Name, "NMEA", Subscribers),

    {noreply, State}.


%% @spec handle_cast(Msg, State) -> {noreply, State}
%%
handle_cast(stop, State) ->
    {stop, normal, State}.


%% @spec terminate(Reason, State) -> ok
%%
%% Close the socket for termination
%%
terminate(_Reason, State) ->
    io:format("~p stopping~n", [?MODULE]),
    #state{socket=Socket} = State,
    ok = gen_tcp:close(Socket).


%% @spec code_change(OldVersion, State, Extra) -> {ok, State}
%%
code_change(_OldVersion, State, _Extra) ->
    {ok, State}.

%% @spec stop(Name) -> {ok}
%%
stop(Name) ->
    gen_server:cast({global, Name}, stop).
